/*
 * Java
 *
 * Copyright 2018-2019 MicroEJ Corp. All rights reserved.
 * For demonstration purpose only.
 * MicroEJ Corp. PROPRIETARY. Use is subject to license terms.
 */
package com.microej.demo.thermorest.cloud;

import java.io.IOException;
import java.net.InetAddress;
import java.util.HashMap;
import java.util.Map;

import org.json.me.JSONException;
import org.json.me.JSONObject;

import com.microej.demo.thermorest.HeadlessThermostat;
import com.microej.demo.thermorest.Messages;
import com.microej.demo.thermorest.model.ThermostatModel;

import ej.bon.Timer;
import ej.bon.TimerTask;
import ej.net.util.NetUtil;
import ej.net.util.NtpUtil;
import ej.util.message.Level;

/**
 * Thermostat publishing changes to a cloud.
 *
 * <ul>
 * <li>It publishes the target to the topic on changes : MicroEJ/thermostat/update/target</li>
 * <li>It publishes the temperature to the topic periodically and on changes :
 * MicroEJ/thermostat/update/temperature</li>
 * <li>It publishes the IP to the topic periodically : MicroEJ/thermostat/ip</li>
 * <li>It subscribes to the topic : MicroEJ/thermostat/set/target</li>
 * </ul>
 */
public abstract class PeriodicPublisherCloud extends CloudConnector {

	private static final int DELAY = 5_000;

	private static final String BASE_TOPIC = "MicroEJ/thermostat/"; //$NON-NLS-1$
	private static final String TOPIC_UPDATE = BASE_TOPIC + "update/"; //$NON-NLS-1$

	/**
	 * Topic used when the temperature is IP address is updated.
	 */
	public static final String TOPIC_IP = BASE_TOPIC + "ip"; //$NON-NLS-1$

	/**
	 * Topic used when the temperature is updated.
	 */
	public static final String TOPIC_TEMPERATURE = TOPIC_UPDATE + "temperature"; //$NON-NLS-1$

	/**
	 * Topic used when the target is updated.
	 */
	public static final String TOPIC_TARGET = TOPIC_UPDATE + "target"; //$NON-NLS-1$

	/**
	 * Topic used to set the target.
	 */
	public static final String TOPIC_SET = BASE_TOPIC + "set/" + TOPIC_TARGET; //$NON-NLS-1$

	private HeadlessThermostat thermostat;
	private TimerTask publish;
	private Timer timer;

	private final Map<String, TimerTask> publishTasks = new HashMap<>();

	private boolean isError;
	private boolean isConnected;

	@Override
	public void start() {
		this.timer = new Timer();
		this.timer.schedule(new TimerTask() {

			@Override
			public void run() {
				Thread.currentThread().setName(getName());
			}
		}, 0);
		this.publish = new TimerTask() {

			@Override
			public void run() {
				publishIP();
				onTemperatureChange(PeriodicPublisherCloud.this.thermostat.getModel().getTemperature());
			}

		};
		this.timer.schedule(this.publish, DELAY, DELAY);
		this.thermostat.addCloudConnector(this);
	}

	/**
	 * Stops the publishing.
	 */
	@Override
	public void stop() {
		TimerTask publish = this.publish;
		this.publish = null;
		if (publish != null) {
			publish.cancel();
		}
		Timer timer = this.timer;
		if (timer != null) {
			timer.cancel();
			this.timer = null;
		}
		this.thermostat.removeCloudConnector(this);
	}

	/**
	 * Sets the thermostat.
	 *
	 * @param thermostat
	 *            the thermostat to set.
	 */
	public void setThermostat(HeadlessThermostat thermostat) {
		this.thermostat = thermostat;
	}

	@Override
	public void onTargetChange(float newTarget) {
		publish(new PublishMessage() {

			@Override
			public String getTopic() {
				return TOPIC_TARGET;
			}

			@Override
			public String getMessage() {
				return Float.toString(PeriodicPublisherCloud.this.thermostat.getModel().getTargetTemperature());
			}
		});
	}

	@Override
	public void onTemperatureChange(float newTemperature) {
		publish(new PublishMessage() {

			@Override
			public String getTopic() {
				return TOPIC_TEMPERATURE;
			}

			@Override
			public String getMessage() {
				return Float.toString(PeriodicPublisherCloud.this.thermostat.getModel().getTemperature());
			}
		});
	}

	@Override
	public boolean isConnected() {
		return this.isConnected;
	}

	/**
	 * Parse a JSON string to update the thermostat.
	 *
	 * @param json
	 *            a JSON with parameters target, temperature and/or unit
	 *            <code>{target=20,temperature=25,unit="°C"}</code>
	 */
	protected void receive(final String json) {
		try {
			JSONObject jsonObject = new JSONObject(json);

			double target = jsonObject.optDouble(ThermostatModel.TARGET);
			if (target != Double.NaN) {
				this.thermostat.getModel().setTargetTemperature((float) target);
			}

			double temperature = jsonObject.optDouble(ThermostatModel.TEMPERATURE);
			if (temperature != Double.NaN) {
				this.thermostat.getModel().setTemperature((float) temperature);
			}

			String unitString = jsonObject.optString(ThermostatModel.UNIT);
			if (unitString != null) {
				if (unitString.length() == 2) {
					if (unitString.charAt(1) == 'C' || unitString.charAt(1) == 'c') {
						unitString = ThermostatModel.CELCIUS;
					} else if (unitString.charAt(1) == 'F' || unitString.charAt(1) == 'f') {
						unitString = ThermostatModel.FARENHEIGHT;
					}
				}
				this.thermostat.getModel().setUnit(unitString);
			}
		} catch (NumberFormatException | JSONException e) {
			Messages.LOGGER.log(Level.INFO, Messages.CATEGORY, Messages.INVALID_PARAMETER, e);
		}
	}

	/**
	 * Sets the cloud connector to be in error state. Notify the listeners of the change.
	 *
	 * @param isError
	 *            <code>true</code> if the connector is in error state.
	 */
	private void setIsError(boolean isError) {
		if (isError != this.isError) {
			this.isError = isError;
			setChanged();
		}
	}

	/**
	 * Sets the cloud connector to be connected. Notify the listeners of the change.
	 *
	 * @param isConnected
	 *            <code>true</code> if the connector is connected.
	 */
	protected void setIsConnected(boolean isConnected) {
		if (isConnected != this.isConnected) {
			this.isConnected = isConnected;
			setChanged();
		}
	}

	/**
	 * Disconnects from the cloud.
	 *
	 * @throws IOException
	 *             if an {@link IOException} occurred.
	 */
	protected void disconnect() throws IOException {
		doDisconnect();
		setIsConnected(false);
	}

	/**
	 * Subscribes to a topic.
	 *
	 * @param topic
	 *            the topic to subscribe to, cannot be <code>null</code>.
	 * @throws IOException
	 *             if an {@link IOException} occurred.
	 */
	protected abstract void subscribe(String topic) throws IOException;

	/**
	 * Publishes to a topic.
	 *
	 * @param topic
	 *            the topic to subscribe to.
	 * @param message
	 *            the message to publish.
	 * @throws IOException
	 *             if an {@link IOException} occurred.
	 */
	protected abstract void doPublish(String topic, String message) throws IOException;

	/**
	 * Disconnects from the cloud.
	 *
	 * @throws IOException
	 *             when the disconnect failed.
	 */
	protected abstract void doDisconnect() throws IOException;

	/**
	 * Connects to the cloud.
	 *
	 * @throws IOException
	 *             when the connect failed.
	 */
	protected abstract void doConnect() throws IOException;

	/**
	 * Request a publication of the IP address.
	 */
	protected void publishIP() {
		InetAddress ip = NetUtil.getFirstHostAddress();
		if (ip != null) {
			final String host = ip.getHostAddress();
			publish(new PublishMessage() {

				@Override
				public String getTopic() {
					return TOPIC_IP;
				}

				@Override
				public String getMessage() {
					return host;
				}
			});
		}
	}

	private void connect() throws IOException {
		doConnect();
		setIsConnected(true);
		publishIP();
		subscribe(TOPIC_SET);
		setIsError(false);
		notifyObservers();
	}

	private void publish(final PublishMessage message) {
		Timer timer = this.timer;
		if (timer != null) {
			final String topic = message.getTopic();
			if (!this.publishTasks.containsKey(topic)) {
				TimerTask task = new TimerTask() {

					@Override
					public void run() {
						if (isReady()) {
							try {
								if (!isConnected()) {
									connect();
								}
								if (isConnected()) {
									doPublish(topic, message.getMessage());
								}
							} catch (IOException e) {
								Messages.LOGGER.log(Level.SEVERE, Messages.CATEGORY, Messages.CLOUD_ERROR, e);
								setIsError(true);
								try {
									disconnect();
								} catch (IOException e1) {
									Messages.LOGGER.log(Level.SEVERE, Messages.CATEGORY, Messages.CLOUD_ERROR, e1);
								}
								notifyObservers();
								try {
									NtpUtil.updateLocalTime();
								} catch (IOException e1) {
									Messages.LOGGER.log(Level.SEVERE, Messages.CATEGORY, Messages.CLOUD_ERROR, e1);
								}
							}
						}
						PeriodicPublisherCloud.this.publishTasks.remove(topic);
					}
				};
				this.publishTasks.put(topic, task);
				timer.schedule(task, 0);
			}

		}

	}

	/**
	 * Checks whether the listener is ready to be connected.
	 *
	 * @return <code>true</code> if the connector is ready to be connected.
	 */
	public boolean isReady() {
		return true;
	}
}
