/*
 * Java
 *
 * Copyright 2018-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.microej.profiling;

import java.util.logging.Level;
import java.util.logging.Logger;

import com.microej.profiling.internal.Strings;
import com.microej.profiling.profiler.InstantProfiler;
import com.microej.profiling.profiler.Profiler;

import ej.bon.Timer;
import ej.bon.TimerTask;
import ej.components.dependencyinjection.ServiceLoaderFactory;

/**
 * Profile an {@link InstantProfiler} using a sample rate and a print rate. If the sampling rate is less than the print,
 * the Profiler will print the min, max and average values gathered in between two prints.
 *
 * By default it will use the components timer and create one if non is found, a timer can be provided using the
 * function {@link #setTimer(Timer)}.
 *
 * The profiler can compute the range from the {@link #start()} see ({@link #watchOverallRange()}; or/and in between two
 * print see {@link #watchIntervalRange()}.
 */
public class AutomaticProfiler {

	private static final int DEFAULT_SAMPLING = 50;
	private static final int DEFAULT_PRINT = 1000;

	private final Object mutex = new Object();
	private int samplingRate = DEFAULT_SAMPLING;
	private int printRate = DEFAULT_PRINT;
	private Timer timer;
	private TimerTask sampleTask;
	private long[] values;
	private int current;
	private long overallMin;
	private long overallMax;
	private long overallAvg;
	private long overAllCount;

	private final InstantProfiler profiler;
	private Level logLevel = Level.INFO;

	private boolean overall;
	private boolean interval;

	/**
	 * Instantiates a {@link AutomaticProfiler}.
	 *
	 * @param profiler
	 *            the profiler to follow.
	 */
	public AutomaticProfiler(InstantProfiler profiler) {
		if (profiler == null) {
			throw new NullPointerException();
		}
		this.overall = true;
		this.interval = false;
		this.profiler = profiler;
	}

	/**
	 * Gets whether the profiler watches the range from the start.
	 *
	 * @return whether the profiler watches the range from the start.
	 */
	public boolean watchOverallRange() {
		return this.overall;
	}

	/**
	 * Sets whether the profiler watches the range from the start.
	 *
	 * @param overall
	 *            whether the profiler watches the range from the start.
	 */
	public void watchOverallRange(boolean overall) {
		this.overall = overall;
	}

	/**
	 * Gets whether the profiler watches the range from the last call.
	 *
	 * @return whether the profiler watches the range from the last call.
	 */
	public boolean watchIntervalRange() {
		return this.interval;
	}

	/**
	 * Sets whether the profiler watches the range from the last call.
	 *
	 * @param interval
	 *            whether the profiler watches the range from the last call.
	 */
	public void watchIntervalRange(boolean interval) {
		this.interval = interval;
	}

	/**
	 * Sets the log level.
	 *
	 * @param logLevel
	 *            the level to use.
	 */
	public void setLogLevel(Level logLevel) {
		this.logLevel = logLevel;
	}

	/**
	 * Gets the sampling rate. The sampling rate is the frequency that the profiler will use to sample values.
	 *
	 * @return the samplingRate.
	 */
	public int getSamplingRate() {
		return this.samplingRate;
	}

	/**
	 * Gets the print rate. The print rate is the frequency that the profiler will use to print the values.
	 *
	 * @return the printRate.
	 */
	public int getPrintRate() {
		return this.printRate;
	}

	/**
	 * Sets the print and sampling rates.
	 * <p>
	 * The print rate is the frequency that the profiler will use to print the values.
	 * <p>
	 * The sampling rate is the frequency that the profiler will use to sample values.
	 * <p>
	 * It print rate be bigger or equals to the sampling rate.
	 *
	 * @param printRate
	 *            the printRate to set should be more than 0.
	 * @param samplingRate
	 *            the samplingRate to set should be more than 0.
	 * @throws IllegalArgumentException
	 *             when printRate &le; 0 or samplingRate &le; 0 or samplingRate &gt; printRate.
	 */
	public void setRates(int printRate, int samplingRate) {
		if (printRate <= 0 || samplingRate <= 0 || samplingRate > printRate) {
			throw new IllegalArgumentException();
		}
		this.printRate = printRate;
		this.samplingRate = samplingRate;
	}

	/**
	 * Sets the timer used for the sampling.
	 *
	 * @param timer
	 *            the timer to use.
	 */
	public void setTimer(Timer timer) {
		this.timer = timer;
	}

	/**
	 * Starts the profiling.
	 */
	public void start() {
		this.overAllCount = 0;
		this.overallMin = Long.MAX_VALUE;
		this.overallMax = Long.MIN_VALUE;
		synchronized (this.mutex) {
			if (this.sampleTask == null) {
				newValues();
				Timer timer = getTimer();

				this.sampleTask = new TimerTask() {

					@Override
					public void run() {
						try {
							sample();
						} catch (NullPointerException t) {
							// samples can become null if stop during a sample.
						}
					}

				};

				timer.schedule(this.sampleTask, 0, this.samplingRate);
			}
		}
	}

	/**
	 * Stops the profiling.
	 */
	public void stop() {
		synchronized (this.mutex) {
			if (this.sampleTask != null) {
				this.sampleTask.cancel();
				this.sampleTask = null;
				this.values = null;
			}
		}
	}

	/**
	 * Gets the profiler.
	 *
	 * @return the profiler.
	 */
	public Profiler getProfiler() {
		return this.profiler;
	}

	/**
	 * Prints the values sampled.
	 *
	 * @param values
	 *            the values.
	 */
	protected void print(long[] values) {
		newValues();
		Profiler profiler = this.profiler;
		StringBuilder out = new StringBuilder();
		Logger logger = Logger.getLogger(profiler.getClass().getSimpleName());
		if (values.length > 1) {
			long min = values[0];
			long max = min;
			long average = 0;
			for (long sample : values) {
				min = Math.min(min, sample);
				max = Math.max(max, sample);
				average += sample;
			}
			this.overallMin = Math.min(min, this.overallMin);
			this.overallMax = Math.max(max, this.overallMax);
			long nextOverallCount = this.overAllCount + values.length;
			this.overallAvg = (this.overallAvg * this.overAllCount + average) / nextOverallCount;
			this.overAllCount = nextOverallCount;

			if (this.interval) {
				print(profiler, Strings.INTERVAL, out, min, max, average / values.length);
			}
			if (this.interval && this.overall) {
				out.append('\n');
			}
			if (this.overall) {
				print(profiler, Strings.OVERALL, out, this.overallMin, this.overallMax, this.overallAvg);
			}
		} else if (values.length == 1) {
			out.append(profiler.printValue(values[0]));
		}
		logger.log(this.logLevel, out.toString());
	}

	private void print(Profiler profiler, String type, StringBuilder out, long min, long max, long average) {
		out.append('\t' + type + '\t' + profiler.getName() + '\t');
		if (profiler.isBiggerPrinted(min, max)) {
			long tmp = min;
			min = max;
			max = tmp;
		}
		out.append(Strings.MIN);
		out.append(profiler.printValue(min));
		out.append(Strings.MAX);
		out.append(profiler.printValue(max));
		out.append(Strings.AVG);
		out.append(profiler.printValue(average));
	}

	private void sample() {
		if (this.current >= this.values.length) {
			print(this.values);
		}
		this.values[this.current] = this.profiler.getCurrentValue();
		this.current++;
	}

	private Timer getTimer() {
		Timer timer = this.timer;
		if (timer == null) {
			timer = ServiceLoaderFactory.getServiceLoader().getService(Timer.class, Timer.class);
		}
		return timer;
	}

	private void newValues() {
		this.values = new long[Math.round((float) this.printRate / this.samplingRate)];
		this.current = 0;
	}
}
