/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.microej.wadapps.management.util;

import java.util.logging.Level;
import java.util.logging.Logger;

import com.microej.wadapps.kernel.ManagementListener;
import com.microej.wadapps.kernel.TrustService;

import ej.bon.Timer;
import ej.bon.TimerTask;
import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.observable.Observer;
import ej.wadapps.admin.Application;
import ej.wadapps.admin.Application.State;
import ej.wadapps.admin.ApplicationLifecycleListener;
import ej.wadapps.admin.ApplicationOperationException;
import ej.wadapps.admin.ApplicationsManager;
import ej.wadapps.app.Activity;
import ej.wadapps.app.BackgroundService;
import ej.wadapps.management.ActivitiesList;
import ej.wadapps.management.ActivitiesScheduler;
import ej.wadapps.management.ApplicationMetadata;
import ej.wadapps.management.ApplicationMetadataProvider;
import ej.wadapps.management.BackgroundServicesList;
import ej.wadapps.management.LaunchersList;

/**
 * Creates and destroys activities when shown and hidden.
 */
public class ActivitiesSchedulerLowMemory implements ActivitiesScheduler, Observer, ApplicationLifecycleListener {

	/**
	 * An helper class that immediately displays the app the user is launching
	 * on S3. It shows the first activity found in this application, does
	 * nothing otherwise.
	 *
	 * The launcher is not displayed when booting.
	 */
	private final class CurrentAppDisplayManager {

		/**
		 * A property set by S3 init. It contains the ID of the application the
		 * user is launching.
		 */
		private static final String S3_USER_CURRENT_APP = "s3.boot.current.app"; //$NON-NLS-1$

		/**
		 * A property set by S3 init. It contains the ID of activities from the
		 * app the user is launching.
		 */
		private static final String S3_USER_CURRENT_ACTIVITIES = "s3.boot.current.activities"; //$NON-NLS-1$

		private final String applicationId;

		private final String activities;

		private boolean hasBeenShown;

		public CurrentAppDisplayManager() {
			this.applicationId = System.getProperty(S3_USER_CURRENT_APP, "");
			this.activities = System.getProperty(S3_USER_CURRENT_ACTIVITIES, "");
		}

		/**
		 * Tells whether there is an activity to display in user’s current
		 * project.
		 *
		 * @return true if there is an activity, false otherwise.
		 */
		public boolean hasActivityToDisplay() {
			return !this.activities.isEmpty();
		}

		/**
		 * Checks activities updates. It will show one of the current project
		 * activity if found and if it has never been scheduled.
		 */
		public void update() {
			if (this.hasBeenShown) {
				// The activity has already been displayed. Nothing to do.
				return;
			}

			Activity[] activities = ActivitiesSchedulerLowMemory.this.activitiesList.getActivities();

			for (Activity activity : activities) {
				if (isCurrentActivity(activity)) {
					show(activity);
					this.hasBeenShown = true;
					break;
				}

			}
		}

		private boolean isCurrentActivity(Activity activity) {
			if (activity == null) {
				return false;
			}

			final ApplicationsManager applicationManager = ServiceLoaderFactory.getServiceLoader()
					.getService(ApplicationsManager.class);

			final Application application = applicationManager.getApplication(activity);
			return (application != null) && this.applicationId.equals(application.getIdentifier())
					&& this.activities.contains(activity.getClass().getName());
		}
	}

	private ActivitiesList activitiesList;
	private Activity currentLauncher;
	private Activity shownActivity;

	private final CurrentAppDisplayManager currentS3Application;

	private final Logger logger = Logger.getLogger(ActivitiesSchedulerLowMemory.class.getName());

	/**
	 *
	 */
	public ActivitiesSchedulerLowMemory() {
		ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class)
				.addApplicationLifecycleListener(this);
		this.currentS3Application = new CurrentAppDisplayManager();
	}

	/**
	 * Sets the activities list to manage.
	 *
	 * @param activitiesList the activitiesList to set.
	 */
	public void setActivitiesList(ActivitiesList activitiesList) {
		ActivitiesList previousActivitiesList = this.activitiesList;
		if (previousActivitiesList != null) {
			previousActivitiesList.deleteObserver(this);
		}
		this.activitiesList = activitiesList;
		activitiesList.addObserver(this);
	}

	@Override
	public void setLauncher(Activity launcher) {
		this.logger.fine("Launcher set to " + getActivityName(launcher)); //$NON-NLS-1$
		if (launcher != null && !this.currentS3Application.hasActivityToDisplay()) {
			show(launcher);
		}
		boolean changed = launcher != this.currentLauncher;
		this.currentLauncher = launcher;
		if (changed) {
			ManagementListener listener = ServiceLoaderFactory.getServiceLoader().getService(ManagementListener.class);
			if (listener != null) {
				listener.launcherChanged(launcher == null ? null : launcher.getID());
			}
		}
	}

	@Override
	public Activity getLauncher() {
		return this.currentLauncher;
	}

	@Override
	public void showLauncher() {
		this.logger.fine("showLauncher()"); //$NON-NLS-1$
		try {
			show(this.currentLauncher);
		} catch (NullPointerException e) {
			this.logger.warning("No more activity to show."); //$NON-NLS-1$
		}
	}

	@Override
	public synchronized void show(Activity activity) throws NullPointerException {
		if (this.logger.isLoggable(Level.FINE)) {
			this.logger.fine("Showing " + getActivityName(activity)); //$NON-NLS-1$
		}
		Activity previousActivity = this.shownActivity;
		boolean logFiner = this.logger.isLoggable(Level.FINER);
		boolean logFinest = this.logger.isLoggable(Level.FINEST);

		// Implementation specific: do not run activity life-cycle if it is
		// already
		// shown
		// need to test that activity is not null, see WI19724
		if (previousActivity == null || activity == null || !previousActivity.equals(activity)) {

			if (previousActivity != null) {
				try {
					if (logFiner) {
						this.logger.finer("Pausing    " + getActivityName(previousActivity)); //$NON-NLS-1$
					}
					previousActivity.onPause();
					if (logFiner) {
						this.logger.finer("Stopping   " + getActivityName(previousActivity)); //$NON-NLS-1$
					}
					previousActivity.onStop();
					if (logFiner) {
						this.logger.finer("Destroying " + getActivityName(previousActivity)); //$NON-NLS-1$
					}
					previousActivity.onDestroy();
					if (logFiner) {
						this.logger.finer("Destroyed " + getActivityName(previousActivity)); //$NON-NLS-1$
					}
				} catch (Throwable t) {
					this.logger.log(Level.SEVERE, "Error hiding activity.", t); //$NON-NLS-1$
					// FIXME what to do?
				}
			}
			this.shownActivity = activity;
			if (activity == null) {
				throw new NullPointerException();
			}
			try {
				if (logFiner) {
					this.logger.finer("Creating " + getActivityName(activity)); //$NON-NLS-1$
				}
				activity.onCreate();
				if (logFiner) {
					this.logger.finer("Starting " + getActivityName(activity)); //$NON-NLS-1$
				}
				activity.onStart();
				if (logFiner) {
					this.logger.finer("Resuming " + getActivityName(activity)); //$NON-NLS-1$
				}
				activity.onResume();
				if (logFiner) {
					this.logger.finer("Resumed  " + getActivityName(activity)); //$NON-NLS-1$
				}
			} catch (Throwable t) {
				this.logger.log(Level.SEVERE, "Error showing activity.", t); //$NON-NLS-1$
				this.shownActivity = null;
			}
			if (previousActivity != null) {
				this.tryStopApplication(previousActivity, this.shownActivity);
			}
		}
		if (logFinest) {
			this.logger.finest("show OUT"); //$NON-NLS-1$
		}
	}

	static String getActivityName(Activity activity) {
		if (activity == null) {
			return null;
		}
		String id = activity.getID();
		if (id != null) {
			return id;
		} else {
			return activity.getClass().getSimpleName() + " from " + getApplicationName(activity); //$NON-NLS-1$
		}
	}

	static String getApplicationName(Activity activity) {
		ApplicationsManager am = ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class);
		if (am != null) {
			Application app = am.getApplication(activity);
			if (app != null) {
				return getApplicationName(app);
			}
		}
		return "<unknown>"; //$NON-NLS-1$
	}

	static String getApplicationName(Application application) {
		ApplicationMetadataProvider metadataProvider = ServiceLoaderFactory.getServiceLoader()
				.getService(ApplicationMetadataProvider.class);
		if (metadataProvider != null) {
			ApplicationMetadata metadata = metadataProvider.getApplicationMetadata(application);
			if (metadata != null) {
				String name = metadata.getApplicationIdentifier();
				if (name != null) {
					return name;
				}
			}
		}
		return application.getIdentifier();
	}

	/**
	 * Tries to stops the specified application if it has no background services
	 * and if it's not a Launcher nor a System application.
	 */
	private void tryStopApplication(Activity fromActivity, Activity toActivity) {
		if (fromActivity != null) {
			ApplicationsManager am = ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class);
			if (am != null) {
				final Application fromApplication = (fromActivity == null) ? null : am.getApplication(fromActivity);
				Application toApplication = (toActivity == null) ? null : am.getApplication(toActivity);
				boolean logFiner = this.logger.isLoggable(Level.FINER);
				if (logFiner) {
					String from = (fromApplication == null) ? null : fromApplication.getIdentifier();
					String to = (toApplication == null) ? null : toApplication.getIdentifier();
					this.logger.finer("app: " + from + " --> " + to); //$NON-NLS-1$
				}
				if (fromApplication != null && !fromApplication.equals(toApplication)) {
					if (!isSystemApp(fromApplication)) {
						if (!isLauncherApp(am, fromApplication)) {
							if (hasService(am, fromApplication)) {
								if (logFiner) {
									this.logger.finer(fromApplication.getIdentifier()
											+ " not stopped, has background services, "); //$NON-NLS-1$
								}
							} else {
								if (logFiner) {
									this.logger.finer("Scheduling stop of " + getApplicationName(fromApplication)); //$NON-NLS-1$
								}
								ServiceLoaderFactory.getServiceLoader().getService(Timer.class)
										.schedule(new TimerTask() {

											@Override
											public void run() {
												try {
													if (fromApplication.getState() == State.STARTED) {
														fromApplication.stop();
													}
												} catch (IllegalStateException e) {
													// ignore, it might already
													// be stopped.
												} catch (SecurityException | ApplicationOperationException e) {
													Logger.getLogger(ActivitiesScheduler.class.getName())
															.log(Level.SEVERE, "", e); //$NON-NLS-1$
												}
											}
										}, 200);
							}
						} else {
							if (logFiner) {
								this.logger.finer(fromApplication.getIdentifier() + " not stopped, is a Launcher"); //$NON-NLS-1$
							}
						}
					} else {
						if (logFiner) {
							this.logger.finer(fromApplication.getIdentifier() + " not stopped, is a SystemApp"); //$NON-NLS-1$
						}
					}
				}
			}

		}
	}

	private boolean isLauncherApp(ApplicationsManager am, Application application) {
		LaunchersList launchersList = ServiceLoaderFactory.getServiceLoader().getService(LaunchersList.class);
		if (launchersList != null) {
			for (Activity activity : launchersList.getLaunchers()) {
				if (am.getApplication(activity).equals(application)) {
					return true;
				}
			}
		}
		return false;
	}

	private boolean hasService(ApplicationsManager am, Application application) {
		BackgroundServicesList backgroundServicesList = ServiceLoaderFactory.getServiceLoader()
				.getService(BackgroundServicesList.class);
		// Do not stop applications with BG services
		if (backgroundServicesList != null) {
			for (BackgroundService backgroundService : backgroundServicesList.getBackgroundServices()) {
				if (am.getApplication(backgroundService).equals(application)) {
					return true;
				}
			}
		}
		return false;
	}

	private boolean isSystemApp(Application application) {
		TrustService trustService = ServiceLoaderFactory.getServiceLoader().getService(TrustService.class);
		if (trustService != null) {
			return trustService.isSystemApp(application);
		}
		return false;
	}

	@Override
	public void hide(Activity activity) {
		if (this.logger.isLoggable(Level.FINE)) {
			this.logger.fine("Hiding: " + getActivityName(activity)); //$NON-NLS-1$
		}
		if (activity.equals(this.shownActivity)) {
			hideShownActivity();
		}
	}

	private void hideShownActivity() {
		if (this.logger.isLoggable(Level.FINER)) {
			this.logger.finer("hideShownActivity()"); //$NON-NLS-1$
		}
		this.showLauncher();
	}

	@Override
	public void update() {
		if (this.logger.isLoggable(Level.FINER)) {
			Activity[] activities = this.activitiesList.getActivities();
			this.logger.finer("Activities list updated."); //$NON-NLS-1$
			this.logger.finer("Activities[" + "] = {"); //$NON-NLS-1$ //$NON-NLS-2$

			for (Activity activity : activities) {
				this.logger.finer("    " + getActivityName(activity) + ","); //$NON-NLS-1$ //$NON-NLS-2$
			}
			this.logger.finer("}"); //$NON-NLS-1$
		}

		this.currentS3Application.update();
	}

	@Override
	public void stateChanged(Application application) {
		if (application.getState().equals(State.INSTALLED)) {
			update();
		}
	}
}
