/*
 * Java
 *
 * Copyright 2014-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.components.dependencyinjection;

import ej.annotation.Nullable;
import ej.components.ServiceProvider;

/**
 * A simple dependency injection facility.
 */
public interface ServiceLoader extends ServiceProvider {

	/**
	 * {@inheritDoc}
	 * <p>
	 * Equivalent to calling {@link #getService(Class, boolean)} passing @{code true} as @{code singleton}.
	 */
	@Override
	@Nullable
	<T> T getService(Class<T> service);

	/**
	 * Gets the registered implementation for the given service. If there is no registered implementation, an instance
	 * of the default implementation is returned.
	 *
	 * @param service
	 *            the service.
	 * @param defaultImplementation
	 *            the default implementation to use.
	 * @param <T>
	 *            the service type.
	 * @return the instance of the given service.
	 */
	<T> T getService(Class<T> service, Class<? extends T> defaultImplementation);

	/**
	 * Gets the registered implementation for the given service.
	 * <p>
	 * If a singleton is asked and if an instance of this service has already been created, this instance is returned.
	 * Otherwise, creates a new instance of the implementation type.
	 *
	 * @param service
	 *            the service.
	 * @param singleton
	 *            if {@code true} returns an already created instance if any, creates a new one otherwise.
	 * @param <T>
	 *            the service type.
	 * @return an instance of the given service or {@code null} if no implementation is available.
	 * @throws ServiceLoadingException
	 *             if an error occurs while instantiating the service implementation.
	 */
	@Nullable
	<T> T getService(Class<T> service, boolean singleton) throws ServiceLoadingException;

	/**
	 * Gets the registered implementation for the given service. If there is no registered implementation, an instance
	 * of the default implementation is returned.
	 * <p>
	 * If a singleton is asked and if an instance of this service has already been created, this instance is returned.
	 * Otherwise, creates a new instance of the implementation type.
	 *
	 * @param service
	 *            the service.
	 * @param defaultImplementation
	 *            the default implementation to use.
	 * @param singleton
	 *            if {@code true} returns an already created instance if any, creates a new one otherwise.
	 * @param <T>
	 *            the service type.
	 * @return an instance of the given service.
	 * @throws ServiceLoadingException
	 *             if an error occurs while instantiating the service implementation.
	 */
	<T> T getService(Class<T> service, Class<? extends T> defaultImplementation, boolean singleton);

}
