/*
 * Copyright 2013-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.components.registry;

/**
 * Manages the life-cycle of a bundle.
 * <p>
 * A bundle is a set of public interfaces that provide access to some specific application functionality or feature.
 */
public interface BundleActivator {

	/**
	 * The bundle states.
	 */
	enum State {
		/**
		 * The bundle is created and ready to be initialized.
		 */
		Created,
		/**
		 * The bundle is initialized and ready to be linked.
		 *
		 * @see BundleActivator#initialize()
		 */
		Initialized,
		/**
		 * The bundle is linked and ready to be started.
		 *
		 * @see BundleActivator#link()
		 */
		Linked,
		/**
		 * The bundle is started.
		 *
		 * @see BundleActivator#start()
		 */
		Started,
		/**
		 * The bundle is stopped.
		 *
		 * @see BundleActivator#stop()
		 */
		Stopped
	}

	/**
	 * Initializes the bundle.
	 * <p>
	 * Initializes and registers bundle services in the {@link BundleRegistry} instance.
	 *
	 * @see BundleRegistry#register(Class, Object)
	 */
	void initialize();

	/**
	 * Prepares the bundle.
	 * <p>
	 * Links bundle services to other services.
	 *
	 * @see BundleRegistry#getService(Class)
	 */
	void link();

	/**
	 * Starts the bundle.
	 * <p>
	 * At this step, all the bundles are prepared and ready. The services could be started.
	 */
	void start();

	/**
	 * Stops the bundle.
	 * <p>
	 * Unlinks it from the others services and clean possible resources and threads.
	 */
	void stop();

}
