/*
 * Java
 *
 * Copyright 2016-2017 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package ej.data;


import java.io.Flushable;
import java.io.IOException;

/**
 * This abstract class provides a default implementation for the {@link DataWriter} interface with the following features:
 * <ul>
 * <li>Implements structured data checks mandatory by the {@link DataWriter} interface (an {@link IllegalStateException} thrown on error).</li>
 * <li>Defines encoding methods with suffix <code>0</code> (e.g <code>writeBoolean</code> calls <code>writeBoolean0</code>). Methods can throw any kind of exceptions that will be wrapped into an {@link IOException}.</li>
 * <li>An additional {@link #end()} method is called when the data is terminated. This method calls the {@link Flushable#flush()} of the {@link Flushable} given to the constructor.</li>
 * </ul>
 */
public abstract class AbstractDataWriter implements DataWriter{

	protected final DataStreamState state;
	/** {@link Flushable} flushed when the end of the data has been written. May be null. */
	protected final Flushable flushable;

	/**
	 * Creates a new {@link AbstractDataWriter} that will use the given {@link Flushable} to notify the end of the data.
	 * The {@link Flushable#flush()} method of the {@link Flushable} is called when end of the data is reached.
	 * @param flushable the {@link Flushable} used to flush. May be null.
	 */
	public AbstractDataWriter(Flushable flushable){
		this.state = new DataStreamState();
		this.flushable = flushable;
	}

	public AbstractDataWriter(){
		this(null);
	}

	@Override
	public final void writeBooleanValue(boolean value) throws IOException{
		state.checkAndUpdateState(DataStreamState.VALUE_BT);
		try{
			writeBooleanValue0(value);
		}
		catch(Throwable e){
			throw new IOException(e);
		}
		checkEndOfData();
	}

	@Override
	public final void writeIntValue(int value) throws IOException{
		state.checkAndUpdateState(DataStreamState.VALUE_BT);
		try{
			writeIntValue0(value);
		}
		catch(Throwable e){
			throw new IOException(e);
		}
		checkEndOfData();
	}

	@Override
	public final void writeLongValue(long value) throws IOException{
		state.checkAndUpdateState(DataStreamState.VALUE_BT);
		try{
			writeLongValue0(value);
		}
		catch(Throwable e){
			throw new IOException(e);
		}
		checkEndOfData();
	}

	@Override
	public final void writeDoubleValue(double value) throws IOException{
		state.checkAndUpdateState(DataStreamState.VALUE_BT);
		try{
			writeDoubleValue0(value);
		}
		catch(Throwable e){
			throw new IOException(e);
		}
		checkEndOfData();
	}

	@Override
	public final void writeStringValue(String value) throws IOException{
		state.checkAndUpdateState(DataStreamState.VALUE_BT);
		try{
			writeStringValue0(value);
		}
		catch(Throwable e){
			throw new IOException(e);
		}
		checkEndOfData();
	}

	@Override
	public final void writeStringKey(String key) throws IOException{
		state.checkAndUpdateState(DataStreamState.KEY_STRING);
		try{
			writeStringKey0(key);
		}
		catch(Throwable e){
			throw new IOException(e);
		}
	}

	@Override
	public final void writeIntKey(int key) throws IOException{
		state.checkAndUpdateState(DataStreamState.KEY_INT);
		try{
			writeIntKey0(key);
		}
		catch(Throwable e){
			throw new IOException(e);
		}
	}


	@Override
	public final void writeArrayValueStart() throws IOException{
		state.checkAndUpdateState(DataStreamState.VALUE_ARRAY);
		try{
			writeArrayValueStart0();
		}
		catch(Throwable e){
			throw new IOException(e);
		}
	}

	@Override
	public final void writeArrayValueEnd() throws IOException{
		state.checkAndUpdateStateEndCollection(DataStreamState.VALUE_ARRAY);
		try{
			writeArrayValueEnd0();
		}
		catch(Throwable e){
			throw new IOException(e);
		}
		checkEndOfData();
	}

	@Override
	public final void writeMapValueStart() throws IOException{
		state.checkAndUpdateState(DataStreamState.VALUE_MAP);
		try{
			writeMapValueStart0();
		}
		catch(Throwable e){
			throw new IOException(e);
		}
	}

	@Override
	public final void writeMapValueEnd() throws IOException{
		state.checkAndUpdateStateEndCollection(DataStreamState.VALUE_MAP);
		try{
			writeMapValueEnd0();
		}
		catch(Throwable e){
			throw new IOException(e);
		}
		checkEndOfData();
	}


	private void checkEndOfData() throws IOException{
		if(state.checkAndUpdateStateEndOfData()){
			end();
		}
	}

	/**
	 * This method is called when the data has been fully written, giving
	 * a "finalization" hook to the implementation.
	 * Calls the {@link Flushable#flush()} method of the {@link Flushable} given
	 * to the constructor.
	 * @throws IOException on I/O error.
	 */
	protected void end() throws IOException{
		if(flushable != null){
			flushable.flush();
		}
	}

	protected abstract void writeBooleanValue0(boolean value) throws Throwable;
	protected abstract void writeIntValue0(int value) throws Throwable;
	protected abstract void writeLongValue0(long value) throws Throwable;
	protected abstract void writeDoubleValue0(double value) throws Throwable;
	protected abstract void writeStringValue0(String value) throws Throwable;
	protected abstract void writeIntKey0(int value) throws Throwable;
	protected abstract void writeStringKey0(String value) throws Throwable;
	protected abstract void writeArrayValueStart0() throws Throwable;
	protected abstract void writeArrayValueEnd0() throws Throwable;
	protected abstract void writeMapValueStart0() throws Throwable;
	protected abstract void writeMapValueEnd0() throws Throwable;

}
