/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style;

import ej.annotation.NonNull;

/**
 * Represents a stylesheet.
 * <p>
 * A stylesheet associates a style to an element.
 * <p>
 * The style of an element is computed from a set a rules. A rule is the combination of a partial style with a selector.
 * <p>
 * A partial style means that all its attributes may not be set.
 * <p>
 * The selector is used to choose on which elements the style is applied based on their Java type, state, etc and a
 * combination of these parameters.
 * <p>
 * Finally, the style of an element computed by the stylesheet is complete and may be built from several rules partial
 * styles.
 * <p>
 * By default, every element has the same style (initial style):
 * <ul>
 * <li>black foreground,</li>
 * <li>white plain background,</li>
 * <li>empty border (black),</li>
 * <li>medium size plain font profile,</li>
 * <li>no padding,</li>
 * <li>no margin,</li>
 * <li>top-left alignment,</li>
 * <li>simple text management.</li>
 * </ul>
 *
 * @see Style#isComplete()
 */
public interface Stylesheet {

	/**
	 * Gets the style for the given element.
	 * <p>
	 * This method should be called in the display thread to avoid concurrency issues.
	 *
	 * @param element
	 *            the element to get the style for.
	 * @return the style applied to the element.
	 */
	@NonNull
	Style getStyle(@NonNull Element element);

	/**
	 * Sets the default style.
	 * <p>
	 * It will be used as the root style of the cascading resolution.
	 * <p>
	 * If the given style is not complete, only its set attributes override the ones in the current default style.
	 * <p>
	 * This method should be called in the display thread to avoid concurrency issues.
	 *
	 * @param style
	 *            the style.
	 * @throws NullPointerException
	 *             if the given style is <code>null</code>.
	 */
	void setDefaultStyle(@NonNull Style style);

	/**
	 * Resets the default style attributes to their initial value.
	 * <p>
	 * This method should be called in the display thread to avoid concurrency issues.
	 */
	void resetDefaultStyle();

	/**
	 * Adds a rule: a style applied to the elements matching a selector.
	 * <p>
	 * If a rule already exists with the same selector, the styles are merged (the attributes of the new one override
	 * the attributes of the previous one).
	 * <p>
	 * This method should be called in the display thread to avoid concurrency issues.
	 *
	 * @param selector
	 *            the selector.
	 * @param style
	 *            the style.
	 * @throws NullPointerException
	 *             if a parameter is <code>null</code>.
	 */
	void addRule(@NonNull Selector selector, @NonNull Style style);

	/**
	 * Gets the style associated with a selector.
	 * <p>
	 * Any modification on the returned style will not affect the style in the stylesheet.
	 * <p>
	 * This method should be called in the display thread to avoid concurrency issues.
	 *
	 * @param selector
	 *            the selector.
	 * @return the style associated with the selector.
	 * @throws NullPointerException
	 *             if the given selector is <code>null</code>.
	 */
	@NonNull
	Style getRuleStyle(@NonNull Selector selector);

	/**
	 * Removes a rule associated with a selector.
	 * <p>
	 * This method should be called in the display thread to avoid concurrency issues.
	 *
	 * @param selector
	 *            the selector.
	 * @throws NullPointerException
	 *             if the given selector is <code>null</code>.
	 */
	void removeRule(@NonNull Selector selector);

	/**
	 * Resets the stylesheet to its initial state.
	 * <p>
	 * This method should be called in the display thread to avoid concurrency issues.
	 */
	void reset();

}
