/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.background;

import ej.microui.display.GraphicsContext;
import ej.microui.display.Image;
import ej.style.container.Rectangle;

/**
 * Draws a repeated image on the background.
 */
public class TiledImageBackground extends AbstractImageBackground {

	private boolean horizontalRepeat;
	private boolean verticalRepeat;

	/**
	 * Creates a tiled image background.
	 */
	public TiledImageBackground() {
		setFillColor(true);
	}

	/**
	 * Creates a tiled image background specifying its attributes.
	 *
	 * @param image
	 *            the background image to set.
	 * @param alignment
	 *            the alignment to set.
	 * @param fillColor
	 *            <code>true</code> to fill remaining area with background color, <code>false</code> otherwise.
	 * @param horizontalRepeat
	 *            <code>true</code> if the image is repeated horizontally, <code>false</code> otherwise.
	 * @param verticalRepeat
	 *            <code>true</code> if the image is repeated vertically, <code>false</code> otherwise.
	 */
	public TiledImageBackground(Image image, int alignment, boolean fillColor, boolean horizontalRepeat,
			boolean verticalRepeat) {
		super(image, alignment, fillColor);
		this.horizontalRepeat = horizontalRepeat;
		this.verticalRepeat = verticalRepeat;
	}

	/**
	 * Gets whether the image is repeated horizontally or not.
	 *
	 * @return <code>true</code> if the image is repeated horizontally, <code>false</code> otherwise.
	 */
	public boolean isHorizontalRepeat() {
		return this.horizontalRepeat;
	}

	/**
	 * Sets whether or not the image is repeated horizontally.
	 *
	 * @param horizontalRepeat
	 *            <code>true</code> if the image is repeated horizontally, <code>false</code> otherwise.
	 */
	public void setHorizontalRepeat(boolean horizontalRepeat) {
		this.horizontalRepeat = horizontalRepeat;
	}

	/**
	 * Gets whether the image is repeated vertically or not.
	 *
	 * @return <code>true</code> if the image is repeated vertically, <code>false</code> otherwise.
	 */
	public boolean isVerticalRepeat() {
		return this.verticalRepeat;
	}

	/**
	 * Sets whether or not the image is repeated horizontally.
	 *
	 * @param verticalRepeat
	 *            <code>true</code> if the image is repeated vertically, <code>false</code> otherwise.
	 */
	public void setVerticalRepeat(boolean verticalRepeat) {
		this.verticalRepeat = verticalRepeat;
	}

	@Override
	protected void drawImage(GraphicsContext g, Rectangle bounds) {
		Image image = getImage();
		if (image != null) {
			int width = bounds.getWidth();
			int height = bounds.getHeight();
			int imageWidth = image.getWidth();
			int imageHeight = image.getHeight();
			boolean horizontalRepeat = isHorizontalRepeat();
			boolean verticalRepeat = isVerticalRepeat();
			int alignment = getAlignment();

			int hrepeat = getRepeatCount(horizontalRepeat, width, imageWidth);
			int vrepeat = getRepeatCount(verticalRepeat, height, imageHeight);

			int xStart;
			int yStart;
			if ((alignment & GraphicsContext.HCENTER) != 0) {
				if (horizontalRepeat) {
					hrepeat++; // one more because both sides can be cut
				}
				xStart = ((width - imageWidth) >> 1) - imageWidth * (hrepeat >> 1);
			} else if ((alignment & GraphicsContext.RIGHT) != 0) {
				xStart = width - imageWidth * hrepeat;
			} else { // left or undefined
				xStart = 0;
			}
			if ((alignment & GraphicsContext.VCENTER) != 0) {
				if (verticalRepeat) {
					vrepeat++; // one more because both sides can be cut
				}
				yStart = ((height - imageHeight) >> 1) - imageHeight * (vrepeat >> 1);
			} else if ((alignment & GraphicsContext.BOTTOM) != 0) {
				yStart = height - imageHeight * vrepeat;
			} else { // top or undefined
				yStart = 0;
			}

			// draw all image repetitions
			int currentX = xStart;
			for (int i = -1; ++i < hrepeat;) {
				int currentY = yStart;
				for (int j = -1; ++j < vrepeat;) {
					g.drawImage(image, currentX, currentY, GraphicsContext.LEFT | GraphicsContext.TOP);
					currentY += imageHeight;
				}
				currentX += imageWidth;
			}
		}
	}

	private int getRepeatCount(boolean repeat, int size, int imageSize) {
		if (repeat) {
			return ceil(size, imageSize);
		} else {
			return 1;
		}
	}

	private int ceil(int x, int y) {
		return 1 + (x - 1) / y;
	}

}
