/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.border;

import ej.annotation.Nullable;
import ej.microui.display.GraphicsContext;
import ej.style.container.Rectangle;

/**
 * Represents and draw a rectangular colored border.
 */
public class ComplexRectangularBorder implements Border {

	private int top;
	private int bottom;
	private int left;
	private int right;

	/**
	 * Creates an empty complex rectangular border (0 pixel thickness).
	 */
	public ComplexRectangularBorder() {
	}

	/**
	 * Creates a complex rectangular border specifying its edges thickness.
	 *
	 * @param top
	 *            the top thickness to set.
	 * @param right
	 *            the right thickness to set.
	 * @param bottom
	 *            the bottom thickness to set.
	 * @param left
	 *            the left thickness to set.
	 */
	public ComplexRectangularBorder(int top, int right, int bottom, int left) {
		this.top = top;
		this.right = right;
		this.bottom = bottom;
		this.left = left;
	}

	/**
	 * Gets the top thickness.
	 *
	 * @return the top thickness.
	 */
	public int getTop() {
		return this.top;
	}

	/**
	 * Sets the top thickness.
	 *
	 * @param thickness
	 *            the thickness to set.
	 */
	public void setTop(int thickness) {
		this.top = thickness;
	}

	/**
	 * Gets the bottom thickness.
	 *
	 * @return the bottom thickness.
	 */
	public int getBottom() {
		return this.bottom;
	}

	/**
	 * Sets the bottom thickness.
	 *
	 * @param thickness
	 *            the thickness to set.
	 */
	public void setBottom(int thickness) {
		this.bottom = thickness;
	}

	/**
	 * Gets the left thickness.
	 *
	 * @return the left thickness.
	 */
	public int getLeft() {
		return this.left;
	}

	/**
	 * Sets the left thickness.
	 *
	 * @param thickness
	 *            the thickness to set.
	 */
	public void setLeft(int thickness) {
		this.left = thickness;
	}

	/**
	 * Gets the right thickness.
	 *
	 * @return the right thickness.
	 */
	public int getRight() {
		return this.right;
	}

	/**
	 * Sets the right thickness.
	 *
	 * @param thickness
	 *            the thickness to set.
	 */
	public void setRight(int thickness) {
		this.right = thickness;
	}

	@Override
	public void wrap(Rectangle rectangle) {
		rectangle.update(-this.left, -this.top, this.left + this.right, this.top + this.bottom);
	}

	@Override
	public void unwrap(Rectangle rectangle) {
		rectangle.update(this.left, this.top, -(this.left + this.right), -(this.top + this.bottom));
	}

	@Override
	public void apply(GraphicsContext g, Rectangle bounds, int color) {
		int width = bounds.getWidth();
		int height = bounds.getHeight();

		g.setColor(color);

		int topThickness = this.top;
		g.fillRect(0, 0, width, topThickness);

		int bottomThickness = this.bottom;
		g.fillRect(0, height - bottomThickness, width, bottomThickness);

		int leftThickness = this.left;
		g.fillRect(0, 0, leftThickness, height);

		int rightThickness = this.right;
		g.fillRect(width - rightThickness, 0, rightThickness, height);

		g.translate(leftThickness, topThickness);
		unwrap(bounds);
		g.clipRect(0, 0, bounds.getWidth(), bounds.getHeight());
	}

	@Override
	public boolean equals(@Nullable Object obj) {
		if (obj instanceof ComplexRectangularBorder) {
			ComplexRectangularBorder border = (ComplexRectangularBorder) obj;
			return this.top == border.top && this.bottom == border.bottom && this.left == border.left
					&& this.right == border.right;
		}
		return false;
	}

	@Override
	public int hashCode() {
		// Since the fields are not final, they can't be used to compute a constant hashcode.
		return 0;
	}

}
