/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.selector;

/**
 * Provides some facilities to compute selectors specificity.
 * <p>
 * The specificity is divided in four parts (defined by CSS2 specification).
 * <ul>
 * <li>count 1 if the declaration is from is a 'style' attribute rather than a rule with a selector, 0 otherwise (= a)</li>
 * <li>count the number of ID attributes in the selector (= b)</li>
 * <li>count the number of other attributes and pseudo-classes in the selector (= c)</li>
 * <li>count the number of element names and pseudo-elements in the selector (= d)</li>
 * </ul>
 * Concatenating the four numbers a-b-c-d (in a number system with a large base) gives the specificity.
 * <p>
 * In this implementation, each number is stored on a byte in a integer.
 */
public class SpecificityHelper {

	/**
	 * The a number shift in the integer.
	 */
	public static final int A_SHIFT = 24;
	/**
	 * The b number shift in the integer.
	 */
	public static final int B_SHIFT = 16;
	/**
	 * The c number shift in the integer.
	 */
	public static final int C_SHIFT = 8;
	/**
	 * The d number shift in the integer.
	 */
	public static final int D_SHIFT = 0;

	// Prevents initialization.
	private SpecificityHelper() {
	}

	/**
	 * Computes the specificity for the given four numbers a, b, c and d.
	 * 
	 * @param a
	 *            the a number.
	 * @param b
	 *            the b number.
	 * @param c
	 *            the c number.
	 * @param d
	 *            the d number.
	 * @return the specificity for the given numbers.
	 */
	public static int getSpecificity(int a, int b, int c, int d) {
		return (a << A_SHIFT) + (b << B_SHIFT) + (c << C_SHIFT) + (d << D_SHIFT);
	}

}
