/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.text;

import java.util.ArrayList;
import java.util.List;

import ej.annotation.Nullable;
import ej.bon.XMath;
import ej.microui.display.Font;
import ej.microui.display.GraphicsContext;
import ej.mwt.MWT;
import ej.style.container.AlignmentHelper;
import ej.style.container.Rectangle;

/**
 * Text manager implementation that manages text on multiple lines.
 */
public class ComplexTextManager implements TextManager {

	// TODO manage text decoration & text indent & text transform

	// Graphics context constants.
	private int lineHeight;

	/**
	 * Creates a complex text manager with default line height.
	 */
	public ComplexTextManager() {
		this.lineHeight = MWT.NONE;
	}

	/**
	 * Creates a complex text manager specifying its line height.
	 *
	 * @param lineHeight
	 *            the height to set.
	 */
	public ComplexTextManager(int lineHeight) {
		this.lineHeight = lineHeight;
	}

	/**
	 * Sets the line height.
	 *
	 * @param lineHeight
	 *            the height to set.
	 */
	public void setLineHeight(int lineHeight) {
		this.lineHeight = lineHeight;
	}

	/**
	 * Gets the line height.
	 *
	 * @return the line height.
	 */
	public int getLineHeight() {
		return this.lineHeight;
	}

	// TODO: handle the justify case.
	@Override
	public void drawText(GraphicsContext g, String text, Font font, int color, Rectangle area, int alignment) {
		int lineHeight = getLineHeight(font);
		int fontHeight = font.getHeight();

		g.setFont(font);
		g.setColor(color);
		// g.setEllipsis(true);
		int areaWidth = area.getWidth();
		int areaHeight = area.getHeight();
		String[] lines = TextHelper.splitStringInLines(text, font, areaWidth);
		int linesCount = lines.length;
		// Can be computed once.
		int y = computeYTop(fontHeight, lineHeight, linesCount, areaHeight, alignment);

		// Compute first and last visible lines.
		int clipY = g.getClipY();
		int startLine = Math.max((clipY - y) / lineHeight, 0);
		int clipHeight = g.getClipHeight();
		int endLine = Math.min((clipY + clipHeight - y) / lineHeight + 1, linesCount);
		y += startLine * lineHeight;

		// Draw visible lines.
		for (int i = startLine; i < endLine; i++) {
			String line = lines[i];
			int x = AlignmentHelper.computeXLeftCorner(font.stringWidth(line), 0, areaWidth, alignment);
			g.drawString(line, x, y, GraphicsContext.LEFT | GraphicsContext.TOP);
			y += lineHeight;
		}
	}

	@Override
	public Rectangle computeContentSize(String text, Font font, Rectangle availableSize) {
		int lineHeight = getLineHeight(font);

		int availableWidth = availableSize.getWidth();
		String[] lines;
		if (availableWidth == MWT.NONE) {
			lines = TextHelper.splitStringInLines(text);
		} else {
			lines = TextHelper.splitStringInLines(text, font, availableWidth);
		}
		int linesCount = lines.length;
		int textHeight = Math.max(linesCount, 1) * lineHeight;

		// Finds the largest line.
		int textWidth = 0;
		for (int i = 0; i < linesCount; i++) {
			textWidth = Math.max(textWidth, font.stringWidth(lines[i]));
		}

		return new Rectangle(0, 0, textWidth, textHeight);
	}

	private int getLineHeight(Font font) {
		if (this.lineHeight != MWT.NONE) {
			return this.lineHeight;
		} else {
			return font.getHeight();
		}
	}

	@Override
	public int getIndex(int x, int y, String text, Font font, Rectangle area, int alignment) {
		int textLength = text.length();
		if (textLength == 0) {
			return 0;
		}

		x -= area.getX();
		y -= area.getY();

		int lineHeight = getLineHeight(font);
		int fontHeight = font.getHeight();
		int areaWidth = area.getWidth();
		int areaHeight = area.getHeight();
		String[] lines = TextHelper.splitStringInLines(text, font, areaWidth);
		int linesCount = lines.length;
		// Shift to top of the text.
		int yTop = computeYTop(fontHeight, lineHeight, linesCount, areaHeight, alignment);
		y -= yTop;

		// Search line and compute its starting index.
		int lineIndex = XMath.limit(y / lineHeight, 0, linesCount - 1);
		int lineFirstIndex = 0;
		for (int i = -1; ++i < lineIndex;) {
			String line = lines[i];
			lineFirstIndex += line.length() + 1; // Plus one is the line breaker (\n or space).
		}
		String line = lines[lineIndex];
		// Shift to left of the line.
		int xLeft = AlignmentHelper.computeXLeftCorner(font.stringWidth(line), 0, areaWidth, alignment);
		x -= xLeft;

		int min = 0;
		int max = line.length();
		int startX = 0;

		// Search the character under the given position.
		// Use dichotomy.
		while (max - min > 1) {
			int half = (max - min) >> 1;
			int halfWidth = font.substringWidth(line, min, half);
			if (startX + halfWidth > x) {
				// Select first part.
				max = min + half;
			} else {
				// Select second part.
				min += half;
				startX += halfWidth;
			}
		}

		// Select the right side of the selected character.
		if ((startX + font.charWidth(text.charAt(min)) / 2) < x) {
			min++;
		}
		return lineFirstIndex + min;
	}

	@Override
	public Rectangle[] getBounds(int startIndex, int endIndex, String text, Font font, Rectangle area, int alignment) {
		int fontHeight = font.getHeight();
		int lineHeight = getLineHeight(font);
		int areaWidth = area.getWidth();
		int areaHeight = area.getHeight();
		if (text.isEmpty()) {
			int xLeft = AlignmentHelper.computeXLeftCorner(0, 0, areaWidth, alignment);
			int yTop = AlignmentHelper.computeYTopCorner(fontHeight, 0, areaHeight, alignment);
			Rectangle rectangle = new Rectangle(xLeft, yTop, 0, fontHeight);
			return new Rectangle[] { rectangle };
		}

		String[] lines = TextHelper.splitStringInLines(text, font, areaWidth);
		int linesCount = lines.length;
		// Compute top text y.
		int yTop = computeYTop(fontHeight, lineHeight, linesCount, areaHeight, alignment);

		List<Rectangle> result = new ArrayList<>(1);

		// Search start index line.
		int currentIndex = 0;
		int linesLength = lines.length;
		int startY = yTop;
		for (int i = -1; ++i < linesLength;) {
			String line = lines[i];
			// Compute left line x.
			int lineWidth = font.stringWidth(line);
			int xLeft = AlignmentHelper.computeXLeftCorner(lineWidth, 0, areaWidth, alignment);
			int lineLength = line.length();
			if (isOnLine(startIndex, currentIndex, lineLength)) {
				int notSelectedWidth = font.substringWidth(line, 0, startIndex - currentIndex);
				// Update values for the current line.
				int startX = xLeft + notSelectedWidth;
				lineWidth -= notSelectedWidth;

				// Search end index line.
				while (true) {
					if (isOnLine(endIndex, currentIndex, lineLength)) {
						int endX = xLeft + font.substringWidth(line, 0, Math.min(endIndex - currentIndex, lineLength));
						Rectangle rectangle = new Rectangle(startX, startY, endX - startX, fontHeight);
						result.add(rectangle);
						break;
					}
					// End is not in this line.
					// Finish the rectangle for this line.
					Rectangle rectangle = new Rectangle(startX, startY, lineWidth, fontHeight);
					result.add(rectangle);

					// Update values for the next line.
					currentIndex += lineLength + 1; // Plus one is the line breaker (\n or space).
					line = lines[++i];
					// Compute left line x.
					lineWidth = font.stringWidth(line);
					xLeft = AlignmentHelper.computeXLeftCorner(lineWidth, 0, areaWidth, alignment);
					lineLength = line.length();
					startX = xLeft;
					startY += lineHeight;
				}
				break;
			}
			currentIndex += lineLength + 1; // Plus one is the line breaker (\n or space).
			startY += lineHeight;
		}

		return result.toArray(new Rectangle[result.size()]);
	}

	@Override
	public boolean equals(@Nullable Object obj) {
		if (obj instanceof ComplexTextManager) {
			ComplexTextManager textManager = (ComplexTextManager) obj;
			return this.lineHeight == textManager.lineHeight;
		}
		return false;
	}

	@Override
	public int hashCode() {
		// Since the fields are not final, they can't be used to compute a constant hashcode.
		return 0;
	}

	private static boolean isOnLine(int index, int currentIndex, int lineLength) {
		return currentIndex + lineLength + 1 >= index; // Plus one is the line breaker (\n or space).
	}

	private static int computeYTop(int fontHeight, int lineHeight, int linesCount, int areaHeight, int alignment) {
		int y = AlignmentHelper.computeYTopCorner(lineHeight * linesCount, 0, areaHeight, alignment);
		return AlignmentHelper.computeYTopCorner(fontHeight, y, lineHeight, alignment);
	}

}
