/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.text;

import ej.microui.display.Font;
import ej.microui.display.GraphicsContext;
import ej.style.container.AlignmentHelper;
import ej.style.container.Rectangle;

/**
 * Text manager implementation that manages text on a single line.
 */
public class SimpleTextManager implements TextManager {

	@Override
	public void drawText(GraphicsContext g, String text, Font font, int color, Rectangle area, int alignment) {
		g.setFont(font);
		g.setColor(color);
		int x = AlignmentHelper.computeXLeftCorner(font.stringWidth(text), 0, area.getWidth(), alignment);
		int y = AlignmentHelper.computeYTopCorner(font.getHeight(), 0, area.getHeight(), alignment);
		g.drawString(text, x, y, GraphicsContext.LEFT | GraphicsContext.TOP);
	}

	@Override
	public Rectangle computeContentSize(String text, Font font, Rectangle availableSize) {
		int textWidth = font.stringWidth(text);
		int textHeight = font.getHeight();
		return new Rectangle(0, 0, textWidth, textHeight);
	}

	@Override
	public int getIndex(int x, int y, String text, Font font, Rectangle area, int alignment) {
		int textLength = text.length();
		if (textLength == 0) {
			return 0;
		}

		x -= area.getX();
		y -= area.getY();

		// Shift to beginning of the text.
		int xLeft = AlignmentHelper.computeXLeftCorner(font.stringWidth(text), 0, area.getWidth(), alignment);
		int yTop = AlignmentHelper.computeYTopCorner(font.getHeight(), 0, area.getHeight(), alignment);
		x -= xLeft;
		y -= yTop;

		int min = 0;
		int max = textLength;
		int startX = 0;

		// Search the character under the given position.
		// Use dichotomy.
		while (max - min > 1) {
			int half = (max - min) >> 1;
			int halfWidth = font.substringWidth(text, min, half);
			if (startX + halfWidth > x) {
				// Select first part.
				max = min + half;
			} else {
				// Select second part.
				min += half;
				startX += halfWidth;
			}
		}

		// Select the right side of the selected character.
		if ((startX + font.charWidth(text.charAt(min)) / 2) < x) {
			min++;
		}
		return min;
	}

	@Override
	public Rectangle[] getBounds(int startIndex, int endIndex, String text, Font font, Rectangle area, int alignment) {
		// Shift to beginning of the text.
		int xLeft = AlignmentHelper.computeXLeftCorner(font.stringWidth(text), 0, area.getWidth(), alignment);
		int yTop = AlignmentHelper.computeYTopCorner(font.getHeight(), 0, area.getHeight(), alignment);
		int startY = yTop;
		int startX = xLeft + getX(startIndex, 0, font, text);
		int endX = xLeft + getX(endIndex, 0, font, text);

		Rectangle rectangle = new Rectangle(startX, startY, endX - startX, font.getHeight());

		return new Rectangle[] { rectangle };
	}

	private int getX(int index, int currentIndex, Font font, String line) {
		if (index < currentIndex) {
			return font.stringWidth(line);
		} else {
			return font.substringWidth(line, 0, index - currentIndex);
		}
	}

}
