/*
 * Java
 *
 * Copyright 2014-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.watchdog.internal;

import java.util.logging.Level;
import java.util.logging.Logger;

import ej.bon.Timer;
import ej.bon.TimerTask;
import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.watchdog.Watchdog;
import ej.watchdog.WatchdogListener;

/**
 * A simple implementation of the watchdog based on a {@link Timer}.
 */
public class SimpleWatchdog implements Watchdog {

	private final long delay;

	private WatchdogListener watchdogListener;
	private TimerTask task;
	private boolean interrupted;
	private boolean fed;
	private final Timer timer;

	/**
	 * Supervises some activity.
	 *
	 * @param delay
	 *            the minimum duration of inactivity before the listener is notified.
	 */
	public SimpleWatchdog(final long delay) {
		this(delay, ServiceLoaderFactory.getServiceLoader().getService(Timer.class, Timer.class));
	}

	/**
	 * Supervises some activity.
	 *
	 * @param delay
	 *            the minimum duration of inactivity before the listener is notified.
	 * @param timer
	 *            the timer to use.
	 * @throws NullPointerException
	 *             when the timer is <code>null</code>.
	 */
	public SimpleWatchdog(final long delay, final Timer timer) throws NullPointerException {
		if (timer == null) {
			throw new NullPointerException();
		}
		this.delay = delay;
		this.timer = timer;
	}

	@Override
	public void setWatchdogListener(final WatchdogListener watchdogListener) {
		this.watchdogListener = watchdogListener;
	}

	@Override
	public synchronized void arm() {
		if (this.task == null) {
			this.interrupted = false;
			this.fed = false;
			this.task = new TimerTask() {
				@Override
				public void run() {
					tick();
				}

			};
			this.timer.schedule(this.task, this.delay, this.delay);
		}
	}

	private void tick() {
		if (!this.interrupted) {
			// The watchdog has not been feed during one complete tick.
			this.interrupted = true;
			this.fed = false;
			try {
				this.watchdogListener.onTrigger();
			} catch (final Throwable t) {
				this.uncaughtException(t);
			}
		}
		if (this.fed) {
			this.interrupted = false;
		}
	}

	@Override
	public synchronized void stop() {
		TimerTask timerTask = this.task;
		this.task = null;
		if (timerTask != null) {
			timerTask.cancel();
		}
	}

	/**
	 * Feeds the watchdog.
	 */
	@Override
	public void feed() {
		this.interrupted = true;
		this.fed = true;
		if (this.task != null) {
			try {
				this.watchdogListener.onFeed();
			} catch (final Throwable t) {
				uncaughtException(t);
			}
		}
	}

	private void uncaughtException(final Throwable t) {
		Logger.getLogger(Watchdog.class.getName()).log(Level.WARNING, t.getMessage(), t);
	}
}
