/*
* Java
*
* Copyright 2015-2019 MicroEJ Corp. All rights reserved.
* This library is provided in source code for use, modification and test, subject to license terms.
* Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
*/
package ej.widget.basic.drawing;

import ej.microui.display.GraphicsContext;
import ej.style.Style;
import ej.style.container.Rectangle;
import ej.style.util.StyleHelper;
import ej.widget.basic.Box;

/**
 * Implementation of a box that represents a check box.
 * <p>
 * The size of a check box is dependent on the font size.
 * <p>
 * This example shows a simple check box:
 *
 * <pre>
 * CheckBox checkbox = new CheckBox();
 * checkbox.setChecked(true);
 * </pre>
 *
 * <img src="../doc-files/checkbox-simple.png" alt="Simple check box.">
 * <p>
 * This example shows a styled check box:
 *
 * <pre>
 * CheckBox checkbox = new CheckBox();
 * checkbox.setChecked(true);
 *
 * EditableStyle checkboxStyle = new EditableStyle();
 * checkboxStyle.setBorder(new SimpleRectangularBorder(2));
 * checkboxStyle.setPadding(new SimpleOutline(2));
 * checkboxStyle.setForegroundColor(Colors.BLUE);
 * checkbox.mergeStyle(checkboxStyle);
 * </pre>
 *
 * <img src="../doc-files/checkbox-styled.png" alt="Styled check box.">
 * <p>
 * This example shows two styled check box, one checked, one unchecked. The checked one border color changes:
 *
 * <pre>
 * CheckBox checkbox = new CheckBox();
 * checkbox.setChecked(true);
 * CheckBox checkbox2 = new CheckBox();
 *
 * Stylesheet stylesheet = StyleHelper.getStylesheet();
 *
 * EditableStyle checkboxStyle = new EditableStyle();
 * checkboxStyle.setBorder(new SimpleRectangularBorder(2));
 * checkboxStyle.setPadding(new SimpleOutline(2));
 * checkboxStyle.setForegroundColor(Colors.BLUE);
 * stylesheet.addRule(new TypeSelector(CheckBox.class), checkboxStyle);
 *
 * EditableStyle checkedCheckboxStyle = new EditableStyle();
 * checkedCheckboxStyle.setBorderColor(Colors.BLUE);
 * stylesheet.addRule(new AndCombinator(new TypeSelector(CheckBox.class), new StateSelector(State.Checked)),
 * 		checkedCheckboxStyle);
 * </pre>
 *
 * <img src="../doc-files/checkbox-styled-checked.png" alt="Styled checked check box.">
 *
 */
public class CheckBox extends Box {

	@Override
	public void renderContent(GraphicsContext g, Style style, Rectangle bounds) {
		if (isChecked()) {
			int width = bounds.getWidth();
			int height = bounds.getHeight();

			int size = Math.min(width, height);
			g.setColor(style.getForegroundColor());
			g.fillRect(0, 0, size, size);
		}
	}

	@Override
	public Rectangle validateContent(Style style, Rectangle availableSize) {
		int size = StyleHelper.getFont(style).getHeight() >> 1;
		return new Rectangle(0, 0, size, size);
	}

}
