/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.widget.basic.drawing;

import ej.microui.display.GraphicsContext;
import ej.microui.display.shape.AntiAliasedShapes;
import ej.style.Style;
import ej.style.container.AlignmentHelper;
import ej.style.container.Rectangle;
import ej.style.util.StyleHelper;
import ej.widget.basic.AbstractSlider;
import ej.widget.model.BoundedRangeModel;
import ej.widget.model.DefaultBoundedRangeModel;

/**
 * Implementation of a slider that can be horizontal or vertical.
 * <p>
 * The size of a slider is dependent on the font size.
 * <p>
 * This example shows a simple slider:
 *
 * <pre>
 * Slider slider = new Slider(0, 100, 50);
 * </pre>
 *
 * <img src="../doc-files/slider-simple.png" alt="Simple slider.">
 * <p>
 * This example shows a styled slider:
 *
 * <pre>
 * Slider slider = new Slider(0, 100, 50);
 *
 * EditableStyle sliderStyle = new EditableStyle();
 * sliderStyle.setForegroundColor(Colors.BLUE);
 * sliderStyle.setBackgroundColor(Colors.SILVER);
 * sliderStyle.setPadding(new SimpleOutline(1));
 * slider.mergeStyle(sliderStyle);
 * </pre>
 *
 * <img src="../doc-files/slider-styled.png" alt="Styled slider.">
 */
public class Slider extends AbstractSlider {

	/**
	 * The orientation of the slider.
	 */
	protected boolean horizontal;

	/**
	 * Creates a horizontal slider with the given model.
	 *
	 * @param model
	 *            the model to use.
	 */
	public Slider(BoundedRangeModel model) {
		super(model);
		this.horizontal = true;

	}

	/**
	 * Creates a horizontal slider with a default bounded range model.
	 *
	 * @param min
	 *            the minimum value of the slider.
	 * @param max
	 *            the maximum value of the slider.
	 * @param initialValue
	 *            the initial value of the slider.
	 * @see DefaultBoundedRangeModel
	 */
	public Slider(int min, int max, int initialValue) {
		super(min, max, initialValue);
		this.horizontal = true;
	}

	/**
	 * Sets the slider orientation: horizontal or vertical.
	 *
	 * @param horizontal
	 *            <code>true</code> to set the slider horizontal, <code>false</code> to set the slider vertical.
	 */
	public void setHorizontal(boolean horizontal) {
		this.horizontal = horizontal;
	}

	@Override
	protected float computePercentComplete(int pointerX, int pointerY) {
		Rectangle contentBounds = getContentBounds();
		int size = getSize(getStyle());
		// The size of the bar considers the size of the cursor (room left on both sides).
		if (this.horizontal) {
			return (float) (getRelativeX(pointerX) - contentBounds.getX() - (size >> 1))
					/ (contentBounds.getWidth() - size);
		} else {
			return (float) (getRelativeY(pointerY) - contentBounds.getY() - (size >> 1))
					/ (contentBounds.getHeight() - size);
		}
	}

	@Override
	public void renderContent(GraphicsContext g, Style style, Rectangle bounds) {
		int width = bounds.getWidth();
		int height = bounds.getHeight();

		int size = getSize(style);
		int barSize = Math.max(size >> 2, 1);

		int alignment = style.getAlignment();

		// The size of the bar considers the size of the cursor (room left on both sides).
		int barStartX;
		int barStartY;
		int barEndX;
		int barEndY;
		int cursorX;
		int cursorY;
		if (this.horizontal) {
			int yTop = AlignmentHelper.computeYTopCorner(size, 0, height, alignment);
			int centerY = yTop + (size >> 1);
			barStartX = size >> 1;
			barEndX = width - (size >> 1);
			barStartY = centerY;
			barEndY = centerY;
			int barWidth = width - size;
			int completeBarWidth = (int) (getPercentComplete() * barWidth);
			cursorX = barStartX + completeBarWidth;
			cursorY = centerY;
		} else {
			int xLeft = AlignmentHelper.computeXLeftCorner(size, 0, width, alignment);
			int centerX = xLeft + (size >> 1);
			barStartY = size >> 1;
			barEndY = height - (size >> 1);
			barStartX = centerX;
			barEndX = centerX;
			int barHeight = height - size;
			int completeBarHeight = (int) (getPercentComplete() * barHeight);
			cursorX = centerX;
			cursorY = barStartY + completeBarHeight;
		}

		int foregroundColor = style.getForegroundColor();
		g.setColor(foregroundColor);

		AntiAliasedShapes antiAliasedShapes = AntiAliasedShapes.Singleton;
		antiAliasedShapes.setFade(1);

		// Draw the bar.
		antiAliasedShapes.setThickness(barSize - 2);
		antiAliasedShapes.drawLine(g, barStartX, barStartY, barEndX, barEndY);

		// Draw the cursor.
		antiAliasedShapes.setThickness(size - 2);
		antiAliasedShapes.drawPoint(g, cursorX, cursorY);
	}

	@Override
	public Rectangle validateContent(Style style, Rectangle availableSize) {
		int size = getSize(style);
		int minimum = getMinimum();
		int maximum = getMaximum();
		int preferredWidth;
		int preferredHeight;
		if (this.horizontal) {
			preferredWidth = size + (maximum - minimum);
			preferredHeight = size;
		} else {
			preferredWidth = size;
			preferredHeight = size + (maximum - minimum);
		}
		return new Rectangle(0, 0, preferredWidth, preferredHeight);
	}

	private int getSize(Style style) {
		return StyleHelper.getFont(style).getHeight();
	}

}
