/*
 * Java
 *
 * Copyright 2013-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.widget.container;

import ej.annotation.Nullable;
import ej.bon.XMath;
import ej.giml.annotation.Element;
import ej.giml.annotation.ElementAttribute;
import ej.giml.annotation.ElementChild;
import ej.giml.annotation.ElementConstructor;
import ej.microui.display.Display;
import ej.mwt.MWT;
import ej.mwt.Widget;
import ej.style.Style;
import ej.style.container.Rectangle;
import ej.widget.StyledComposite;
import ej.widget.animation.AnimationListener;
import ej.widget.animation.AnimationListenerRegistry;
import ej.widget.basic.drawing.Scrollbar;
import ej.widget.util.SwipeEventHandler;
import ej.widget.util.Swipeable;

/**
 * Allows to scroll a widget horizontally or vertically.
 * <p>
 * Horizontal:
 * <p>
 * <img src="doc-files/scroll_horizontal.png" alt="Horizontal scroll.">
 * <p>
 * Vertical:
 * <p>
 * <img src="doc-files/scroll_vertical.png" alt="Vertical scroll.">
 */
@Element
public class Scroll extends StyledComposite {

	@Nullable
	private Widget content;
	@Nullable
	private Scrollable scrollContent;
	private final Scrollbar scrollbar;
	private boolean showScrollbar;
	private boolean horizontal;

	private boolean needScrollbar;

	// Swipe management.
	@Nullable
	private SwipeEventHandler swipeEventHandler;
	@Nullable
	private AnimationListener animationListener;
	@Nullable
	private final SetShift setShift;
	private int value;
	private boolean shown;
	private boolean shifting;

	/**
	 * Creates a horizontal scroll composite with a visible scrollbar.
	 */
	public Scroll() {
		this(true, true);
	}

	/**
	 * Creates a scroll composite specifying its orientation and the visibility of the scrollbar.
	 *
	 * @param horizontal
	 *            <code>true</code> to scroll horizontally, <code>false</code> to scroll vertically.
	 * @param showScrollbar
	 *            <code>true</code> to show the scrollbar, <code>false</code> otherwise.
	 */
	@ElementConstructor
	public Scroll(@ElementAttribute(defaultValue = "true") boolean horizontal,
			@ElementAttribute(defaultValue = "true") boolean showScrollbar) {
		this.horizontal = horizontal;
		this.scrollbar = new Scrollbar(0);
		this.scrollbar.setHorizontal(horizontal);
		this.showScrollbar = showScrollbar;
		this.setShift = new SetShift();
	}

	private void updateContent(@Nullable Widget oldContent, @Nullable Widget content, boolean showScrollbar) {
		remove(this.scrollbar);

		// Add the scrollbar after the content to make sure it is visible.
		if (oldContent != null) {
			remove(oldContent);
		}
		if (content != null) {
			add(content);
		}
		if (showScrollbar) {
			add(this.scrollbar);
		}
	}

	/**
	 * Sets the content to scroll.
	 *
	 * @param content
	 *            the widget to scroll.
	 */
	@ElementChild
	public void setWidget(Widget content) {
		updateContent(this.content, content, this.showScrollbar);
		this.content = content;
		if (this.content instanceof Scrollable) {
			this.scrollContent = (Scrollable) this.content;
		} else {
			this.scrollContent = null;
		}
	}

	/**
	 * Sets the scroll orientation: horizontal or vertical.
	 *
	 * @param horizontal
	 *            <code>true</code> to scroll horizontally, <code>false</code> to scroll vertically.
	 */
	public void setHorizontal(boolean horizontal) {
		this.horizontal = horizontal;
		this.scrollbar.setHorizontal(horizontal);
		invalidate();
	}

	/**
	 * Sets whether the scrollbar is visible or not.
	 *
	 * @param show
	 *            <code>true</code> to show the scrollbar, <code>false</code> to hide it.
	 */
	public void showScrollbar(boolean show) {
		if (show != this.showScrollbar) {
			this.showScrollbar = show;
			updateContent(this.content, this.content, show);
		}
	}

	@Override
	public Rectangle validateContent(Style style, Rectangle bounds) {
		int widthHint = bounds.getWidth();
		int heightHint = bounds.getHeight();

		// If size is not yet defined -> compute it!
		boolean computeWidth = widthHint == MWT.NONE;
		boolean computeHeight = heightHint == MWT.NONE;

		Widget content = this.content;
		if (content != null) {
			Scrollbar scrollbar = this.scrollbar;

			if (this.horizontal) {
				content.validate(MWT.NONE, heightHint);
				scrollbar.validate(widthHint, MWT.NONE);
			} else {
				content.validate(widthHint, MWT.NONE);
				scrollbar.validate(MWT.NONE, heightHint);
			}

			int contentWidth = content.getPreferredWidth();
			int contentHeight = content.getPreferredHeight();

			if (computeWidth) {
				widthHint = contentWidth;
			}
			if (computeHeight) {
				heightHint = contentHeight;
			}
		}

		setPreferredSize(widthHint, heightHint);

		// Set composite preferred size.
		return new Rectangle(0, 0, widthHint, heightHint);
	}

	@Override
	protected void setBoundsContent(Rectangle bounds) {
		int width = bounds.getWidth();
		int height = bounds.getHeight();

		Widget content = this.content;
		int contentWidth;
		int contentHeight;
		if (content != null) {
			contentWidth = content.getPreferredWidth();
			contentHeight = content.getPreferredHeight();
			content.setSize(contentWidth, contentHeight);
		} else {
			contentWidth = 0;
			contentHeight = 0;
		}

		int excess;
		if (this.horizontal) {
			excess = contentWidth - width;
			this.needScrollbar = excess > 0;
			if (this.needScrollbar) {
				this.scrollbar.setMaximum(excess);
				int x = bounds.getX();
				int scrollbarHeight = this.scrollbar.getPreferredHeight();
				this.scrollbar.setBounds(x, height - scrollbarHeight, width, scrollbarHeight);
			}
		} else {
			excess = contentHeight - height;
			this.needScrollbar = excess > 0;
			if (this.needScrollbar) {
				this.scrollbar.setMaximum(excess);
				int y = bounds.getY();
				int scrollbarWidth = this.scrollbar.getPreferredWidth();
				this.scrollbar.setBounds(width - scrollbarWidth, y, scrollbarWidth, height);
			}
		}
		if (excess > 0) {
			this.swipeEventHandler = new SwipeEventHandler(excess, false, this.horizontal, new Swipeable() {
				@Override
				public void onMove(int currentStep) {
					setShift(currentStep);
				}
			});
			this.swipeEventHandler.moveTo(this.value);
		}

		int contentCoordinate = -this.scrollbar.getValue();
		updateViewport(contentCoordinate);
	}

	@Override
	public void showNotify() {
		super.showNotify();

		this.animationListener = new AnimationListener() {
			@Override
			public void onStartAnimation() {
				Scroll.this.scrollbar.show();
			}

			@Override
			public void onStopAnimation() {
				Scroll.this.scrollbar.hide();
			}
		};
		AnimationListenerRegistry.register(this.animationListener);

		this.shown = true;
	}

	@Override
	public void hideNotify() {
		super.hideNotify();

		AnimationListenerRegistry.unregister(this.animationListener);

		this.shown = false;
		SwipeEventHandler swipeEventHandler = this.swipeEventHandler;
		if (swipeEventHandler != null) {
			swipeEventHandler.stop();
		}
	}

	/**
	 * Scrolls to a position.
	 *
	 * @param position
	 *            the x or y target (depending on the orientation).
	 * @param animate
	 *            whether the scrolling action should be animated.
	 */
	public void scrollTo(int position, boolean animate) {
		Widget content = this.content;
		if (content != null) {
			int max;
			if (this.horizontal) {
				max = content.getWidth() - getWidth();
			} else {
				max = content.getHeight() - getHeight();
			}
			max = Math.max(0, max);
			position = XMath.limit(position, 0, max);
			this.scrollbar.setValue(position);
			if (this.swipeEventHandler != null && animate) {
				this.swipeEventHandler.moveTo(position, SwipeEventHandler.DEFAULT_DURATION);
			} else {
				setShift(position);
			}
		}
	}

	/**
	 * Scrolls to a position without animation.
	 *
	 * @param position
	 *            the x or y target (depending on the orientation).
	 */
	public void scrollTo(int position) {
		scrollTo(position, false);
	}

	@Override
	public boolean handleEvent(int event) {
		SwipeEventHandler swipeEventHandler = this.swipeEventHandler;
		if (swipeEventHandler != null && swipeEventHandler.handleEvent(event)) {
			return true;
		}
		return super.handleEvent(event);
	}

	private void updateViewport(int x, int y, int width, int height) {
		Scrollable scrollContent = this.scrollContent;
		if (scrollContent != null) {
			scrollContent.updateViewport(x, y, width, height);
		}
		Widget content = this.content;
		if (content != null) {
			content.setLocation(x, y);
		}
	}

	private void updateViewport(int contentCoordinate) {
		Widget content = this.content;
		if (content != null) {
			if (this.horizontal) {
				updateViewport(contentCoordinate, content.getY(), getWidth(), getHeight());
			} else {
				updateViewport(content.getX(), contentCoordinate, getWidth(), getHeight());
			}
		}
	}

	private void setShift(int value) {
		this.value = value;
		if (!this.shifting) {
			this.shifting = true;
			Display.getDefaultDisplay().callSerially(this.setShift);
			repaint();
		}
	}

	class SetShift implements Runnable {

		@Override
		public void run() {
			Scroll scroll = Scroll.this;
			scroll.shifting = false;
			if (scroll.shown) {
				scroll.scrollbar.setValue(scroll.value);
				int contentCoordinate = (-scroll.value - scroll.scrollbar.getValue()) / 2;
				updateViewport(contentCoordinate);
			}
		}
	}

}
