/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.widget.navigation;

import java.util.HashMap;
import java.util.Map;

import ej.microui.display.GraphicsContext;
import ej.microui.event.EventGenerator;
import ej.microui.util.EventHandler;
import ej.style.Style;
import ej.style.container.Rectangle;
import ej.style.util.StyleHelper;
import ej.widget.navigation.page.Page;

/**
 * Animates the transition between two pages of a navigator.
 *
 * @see Navigator
 * @deprecated see {@link ej.widget.container.transition.TransitionContainer}.
 */
@Deprecated
public abstract class TransitionManager {

	private Navigator navigator;

	private final Map<EventGenerator, EventHandler> eventHandlers;

	/**
	 * Creates a transition manager.
	 */
	public TransitionManager() {
		this.eventHandlers = new HashMap<>(EventGenerator.get(EventGenerator.class).length);
	}

	/**
	 * Sets the navigator.
	 *
	 * @param navigator
	 *            the navigator to set.
	 * @throws IllegalArgumentException
	 *             if it is already associated with a navigator.
	 */
	void setNavigator(Navigator navigator) {
		if (navigator != null && this.navigator != null) {
			throw new IllegalArgumentException();
		}
		this.navigator = navigator;
	}

	/**
	 * Gets the navigator.
	 *
	 * @return the navigator.
	 */
	protected Navigator getNavigator() {
		return this.navigator;
	}

	/**
	 * Notifies that the transition is started.
	 */
	protected void startTransition() {
		EventGenerator[] eventGenerators = EventGenerator.get(EventGenerator.class);
		for (EventGenerator eventGenerator : eventGenerators) {
			EventHandler eventHandler;
			try {
				eventHandler = eventGenerator.getEventHandler();
			} catch (Throwable e) { // catch (SecurityException e) {
				// Cannot retrieve this event generator handler…
				// MicroUI should check a security permission. Currently, the VM throw an IllegalAccessError F3.
				continue;
			}
			this.eventHandlers.put(eventGenerator, eventHandler);
			try {
				eventGenerator.setEventHandler(null);
			} catch (SecurityException e) {
				// Cannot change this event generator handler… forget it!.
				this.eventHandlers.remove(eventGenerator);
			}
		}
	}

	/**
	 * Notifies that the transition is stopped.
	 */
	protected void stopTransition() {
		EventGenerator[] eventGenerators = EventGenerator.get(EventGenerator.class);
		for (EventGenerator eventGenerator : eventGenerators) {
			EventHandler eventHandler = this.eventHandlers.remove(eventGenerator);
			if (eventHandler != null) {
				try {
					eventGenerator.setEventHandler(eventHandler);
				} catch (SecurityException e) {
					// Should not occur since the permission is checked by startTransition.
					throw new AssertionError(e);
				}
			}
		}
	}

	/**
	 * Animates the transition between two pages of the navigation container.
	 *
	 * @param oldPage
	 *            the old page.
	 * @param newPage
	 *            the new page.
	 * @param forward
	 *            <code>true</code> if going to a page that is after the current one, <code>false</code> otherwise.
	 */
	protected abstract void animate(Page oldPage, Page newPage, boolean forward);

	/**
	 * Renders the content of the navigation container without the border, margin and padding specified in the style.
	 * <p>
	 * Do nothing by default.
	 *
	 * @param g
	 *            the graphics context where to render the content of the renderable.
	 * @param style
	 *            the style to use.
	 * @param bounds
	 *            the remaining size to render the content.
	 */
	protected void render(GraphicsContext g, Style style, Rectangle bounds) {
		// Do nothing by default.
	}

	/**
	 * Handles an event that occurs on the navigation container.
	 *
	 * @param event
	 *            the event to handle.
	 * @return <code>true</code> if the widget has consumed the event, <code>false</code> otherwise.
	 */
	protected boolean handleEvent(int event) {
		return false;
	}

	/**
	 * Adds a page in the navigation container.
	 *
	 * @param page
	 *            the page to add.
	 */
	protected void addPage(Page page) {
		this.navigator.add(page);
	}

	/**
	 * Removes a page from the navigation container.
	 *
	 * @param page
	 *            the page to remove.
	 */
	protected void removePage(Page page) {
		this.navigator.remove(page);
	}

	/**
	 * Sets the current page.
	 *
	 * @param newPage
	 *            the new page.
	 */
	protected void setCurrentPage(Page newPage) {
		this.navigator.setCurrentPage(newPage);
	}

	/**
	 * Gets whether or not it is possible to go backward.
	 *
	 * @return <code>true</code> if it is possible to go backward, <code>false</code> otherwise.
	 */
	protected boolean canGoBackward() {
		return this.navigator.canGoBackward();
	}

	/**
	 * Gets the previous page.
	 *
	 * @return the previous page or <code>null</code> if cannot go backward.
	 */
	protected Page getPreviousPage() {
		return this.navigator.getPreviousPage();
	}

	/**
	 * Shows the previous page.
	 * <p>
	 * Nothing is done if cannot go backward.
	 */
	protected void goBackward() {
		this.navigator.goBackward();
	}

	/**
	 * Gets whether or not it is possible to go forward.
	 *
	 * @return <code>true</code> if it is possible to go forward, <code>false</code> otherwise.
	 */
	protected boolean canGoForward() {
		return this.navigator.canGoForward();
	}

	/**
	 * Gets the next page.
	 *
	 * @return the next page or <code>null</code> if cannot go forward.
	 */
	protected Page getNextPage() {
		return this.navigator.getNextPage();
	}

	/**
	 * Shows the next page.
	 * <p>
	 * Nothing is done if cannot go forward.
	 */
	protected void goForward() {
		this.navigator.goForward();
	}

	/**
	 * Sets the bounds of a new page.
	 *
	 * @param newPage
	 *            the new page.
	 * @see #getContentBounds()
	 */
	protected void setChildBounds(Page newPage) {
		this.navigator.setChildBounds(newPage);
	}

	/**
	 * Gets the bounds of a child page.
	 *
	 * @return the bounds of a child page.
	 */
	protected Rectangle getContentBounds() {
		Navigator navigator = this.navigator;
		Rectangle bounds = new Rectangle(0, 0, navigator.getWidth(), navigator.getHeight());
		Style style = navigator.getStyle();
		Rectangle contentBounds = StyleHelper.computeContentBounds(bounds, style);
		return contentBounds;
	}

	/**
	 * Shows a new page. If there is already a page shown, it is hidden.
	 *
	 * @param newPage
	 *            the new page to show.
	 * @param forward
	 *            <code>true</code> if going to a new page, <code>false</code> if going back in the stack of pages.
	 */
	protected void show(Page newPage, boolean forward) {
		this.navigator.show(newPage, forward);
	}

}
